import torch
#torch.autograd.set_detect_anomaly(True)
from typing import Literal
from surecr.solver import Solver
import surecr.divergence as div_lib
import time

class SURE:
    def __init__(self,
                 variance: float,
                 solver: Solver):
        self._variance = variance
        self._solver = solver
        self._divergence_strategy: \
                Literal['default', 'exact', 'hutchinson', 'hutch++'] = 'default'
        self._solution = None

    @property
    def solution(self):
        """
        Returns solver.solve(y) from the last compute call.
        """
        return self._solution

    def runtimes(self):
        """
        Returns how long it took for the solver to run and how long it took
        the divergence estimator to run during the last compute call.
        """
        return {'solver':
                    self._t_f_solver - self._t_i_solver,
                'divergence':
                    self._t_f_div - self._t_i_div,
                }

    def compute(self, data: torch.Tensor, divergence_parameters={}) -> torch.Tensor:
        """
        Computes and returns SURE for the estimator computed by the solver
        at the point y.

        Currently, divergence_parameters can contain the key "m" to indicate
        how many samples to use during the divergence estimation (which
        dominates the runtime at high dimensions). The default is for m to be
        102.

        In the future we may switch to A-Hutch++ and may change what options
        the divergence_parameters specifies.
        """

        shape = data.shape
        shape_ereased_data = data.reshape((-1,))
        shape_ereased_data.requires_grad_(True)

        def mu_hat(self, y):
            y = y.reshape(shape)
            self._solution = self._solver.solve(y)
            return self._solver.estimate(self._solution).reshape((-1,))

        dim_var_term = -data.numel() * self._variance

        self._t_i_solver = time.monotonic()
        mu_hat_evaled = mu_hat(self, shape_ereased_data)
        self._t_f_solver = time.monotonic()
        norm_squared_term = (torch.norm(mu_hat_evaled - shape_ereased_data)**2).detach()
        self._norm_squared_term = norm_squared_term

        self._t_i_div = time.monotonic()
        if  self._divergence_strategy == 'exact':
            divergence = div_lib.exact_divergence(mu_hat_evaled, shape_ereased_data,
                    **divergence_parameters)
        elif self._divergence_strategy == 'hutchinson':
            divergence = div_lib.hutchinson(mu_hat_evaled, shape_ereased_data,
                     **divergence_parameters)

        elif self._divergence_strategy == 'default' \
                or self._divergence_strategy == 'hutch++':
            divergence = div_lib.hutch_pp(mu_hat_evaled, shape_ereased_data,
                     **divergence_parameters)
        else:
            raise RuntimeError("Unknown divergence strategy.")
        self._t_f_div = time.monotonic()
        self._divergence = divergence

        div_term = 2 * self._variance * divergence
        return dim_var_term + norm_squared_term + div_term



from __future__ import annotations

import torch
from typing import Callable 

"""
This module contains various ways to compute divergences.

Its API is unstable and internal to sure.


Reference paper: https://arxiv.org/pdf/2010.09649.pdf
"""

def vjp_oracle_factory(solution, input_vec):
    def oracle(z, retain_graph=True):
        solution.backward(gradient=z, retain_graph=retain_graph)
        grad = input_vec.grad
        input_vec.grad = None
        return grad
    return oracle

def hutchinson(solution: torch.Tensor, input_vec: torch.Tensor, m: int=1000):
    if solution.numel() <= m:
        return exact_divergence(solution, input_vec)
    vjp = vjp_oracle_factory(solution, input_vec)
    total = 0
    for _ in range(m):
        z = (2 * torch.randint(2, size=solution.shape, device=solution.device) - 1).float()
        total = total + (z * vjp(z)).sum()
    return total / m



def hutch_pp(solution: torch.Tensor, input_vec: torch.Tensor, m: int=102):
    """
    Algorithm taken from reference paper above on Hutch++.
    """
    assert m % 3 == 0
    if solution.numel() <= m:
        return exact_divergence(solution, input_vec)
    vjp = vjp_oracle_factory(solution, input_vec)
    k = m // 3
    S = 2.0 * torch.randint(0, 2, (*solution.shape, k), device=input_vec.device) - 1.0
    G = 2.0 * torch.randint(0, 2, (*solution.shape, k), device=input_vec.device) - 1.0
    BS = torch.empty_like(S)
    for i in range(k):
        BS[:, i] = vjp(S[:,i])
 
    Q, _ = torch.linalg.qr(BS)
    G_prime = G - Q @ (Q.T @ G)
    total = 0
    for i in range(Q.shape[1]):
        z = Q[:,i]
        total = total + (z * vjp(z)).sum()

    for i in range(k - 1):
        z = G_prime[:,i]
        total = total + (z * vjp(z)).sum() / k
    z = G_prime[:,-1]
    total = total + (z * vjp(z, False)).sum() / k

    return total

def exact_divergence(solution: torch.Tensor, input_vec: torch.Tensor):
    divergence = 0
    for (i, gi) in enumerate(solution):
        gi.backward(retain_graph=True)
        divergence = divergence + input_vec.grad[i]
        input_vec.grad = None
    return divergence



import torch
import surecr.admm.admm as admm
import surecr.prox_grad.prox_grad as prox_grad

class Solver:
    """
    For any given instance s of a solver class, s.estimate(s.solve(y)) must be
    differentiable via torch's backpropagation.
    """
    def __init__(self):
        raise NotImplementedError()
    def solve(self, y):
        """
        Returns intermediate value used to estimate the mean of the distribution
        y is sampled from.
        """
        raise NotImplementedError()
    def estimate(self, beta):
        """
        Given the output of a solve call, returns the estimate of the mean of the
        distribution y was sampled from.
        """
        return NotImplementedError()

class CVXPYSolver(Solver):
    def __init__(self, problem, y_parameter, variables, estimate):
        """
        problem must be a CVXPY problem with a single paremeter, y_parameter,
            and variables y_variable.
    
        estimate must be function which takes tensors with values for each variable
            and returns the estimate.

        WARNING: This solver has poor performance on large problems, and can
        have undetected poor accuracy on some moderately-sized problems.
        """
        try:
            from cvxpylayers.torch import CvxpyLayer
        except ImportError:
            raise RuntimeError("Could not import cvxpylayers. Please ensure it is"
                               " correctly installed. Note: cvxpylayers is not"
                               "automatically installed when using conda.")
        self._problem = problem
        self._y_parameter = y_parameter
        self._variables = variables
        self._layer = CvxpyLayer(problem, parameters=[y_parameter], variables=variables)
        self._estimate = estimate

    def solve(self, y):
        return self._layer.forward(y)[0]
    
    def estimate(self, beta):
        return self._estimate(beta)

class ADMMSolver(Solver):
    """
    This solver solves problems of the form with a variant on ADMM:
            min. 1/2 ||A b - y||_2^2 + r(b)
    and estimates the mean of y with A b^* where b^* is the optimal b.

    A is a linear operator defined using <https://github.com/cvxgrp/torch_linops>

    prox_R is a differentiable-with-respect-to-its-first-argument function to
        find the optimal point b for a (v, t) pair of
            min. t r(b) + 1/2 ||b - v||_2^2

    x0 is the point where we begin iterations, it must be chosen
        indepentently of y.

    max_iters, eps_rel, eps_abs control when iterations stop.
    """
    def __init__(self, A, prox_R, x0, device=None, **parameters):
        self._A = A
        self._prox_R = prox_R
        self._x0 = x0
        self._device = device
        self._iters = None
        self._parameters = parameters

    def solve(self, y):
        x, z, _, self._iters = admm.admm(A=self._A, b=y,
                                         prox_R=self._prox_R,
                                         x0=self._x0,
                                         device=self._device,
                                         **self._parameters)
        return (x + z) / 2
    
    def estimate(self, parameters):
        return self._A @ parameters

class ISTASolver(Solver):
    _prox_grad_solver = lambda _, *args, **kwargs: prox_grad.prox_grad(*args, **kwargs)

    def __init__(self, A, prox_R, x0, device=None,
                 lipschitz_iterations=20,
                 lipschitz_vec=None, **parameters):
        self._A = A
        self._prox_R = prox_R
        self._x0 = x0
        self._device = device
        self._iters = None
        self._lipschitz = _compute_lipschitz(
                A,
                torch.ones_like(x0.reshape(-1)) / x0.numel() \
                        if lipschitz_vec is None else lipschitz_vec,
                lipschitz_iterations)
        self._parameters = parameters

    def solve(self, y):
        self._y = y
        self._ATb = self._A.T @ y
        x, self._iters = self._prox_grad_solver(self._f,
                                                self._grad_f,
                                                self._prox_R,
                                                self._x0.reshape(-1),
                                                self._lipschitz,
                                                **self._parameters)
        return x.reshape(self._x0.shape)

    def estimate(self, beta):
        return self._A @ beta

    def _f(self, x):
        return 1/2 * torch.linalg.vector_norm(self._A @ x - self._y)**2

    def _grad_f(self, x):
        """ f(x) = 12 ||Ax -  b||_2^2 = 1/2 (A x - b)^T (Ax - b) =
            1/2 (x^T A^T A x - b^T Ax - x^T A^T b + ||b||_2^2) = 
            1/2 x^T A^T A x - x^T A^T b + 1/2 ||b||_2^2

            grad f(x) = A^T A x -  A^T b
        """
        return self._A.T @ (self._A @ x) - self._ATb

class FISTASolver(ISTASolver):
    """
    This solver solves problems of the form with a variant on FISTA:
            min. 1/2 ||A b - y||_2^2 + r(b)
    and estimates the mean of y with A b^* where b^* is the optimal b.

    A is a linear operator defined using <https://github.com/cvxgrp/torch_linops>

    prox_R is a differentiable-with-respect-to-its-first-argument function to
        find the optimal point b for a (v, t) pair of
            min. t r(b) + 1/2 ||b - v||_2^2

    x0 is the point where we begin iterations, it must be chosen
        indepentently of y.

    lipschitz_iterations is how many iterations of the power method to use
    to approximate the largest eigenvalue of A^T A

    lipschitz_vec is the vector to start the power method. By default, a
    vector of all 1s is used. If this vector is orthogonal to the largest
    eigenvector of A^T A, this argument is mandatory.

    max_iters, eps control when iterations stop.
    """

    #_prox_grad_solver = lambda _, *args: prox_grad.accel_prox_grad_w_linesearch(*args)
    _prox_grad_solver = lambda _, *args, **kwargs: prox_grad.accel_prox_grad(*args, **kwargs)

def _compute_lipschitz(A, x, iters):
    """
    Finds Lipschitz constant of _grad f(x) = A^T A x -  A^T b

    ||A^T A x - A^Tb - A^T A y + A^T b||_2 =
        ||A^T A x - A^T A y||_2 <= ||A^T A|| ||x - y||_2
    Where ||A^T A|| = sigma_1(A^T A), the maximal singular value of A^T A
    """
    assert torch.linalg.vector_norm(x) > 0
    norm_x = float('inf')

    for _ in range(iters):
        x = A.T @ (A @ x)
        norm_x = torch.linalg.vector_norm(x)
        x /= norm_x
        if torch.abs(norm_x).item() <= 1e-6:
            raise RuntimeError("Either uniform vector or provided "
                               "lipschitz vector is in A^T A's null space. "
                               "Please provide a vector which is NOT "
                               "orthogonal to A^T A's largest eigenvector.")
    #print(f"Lipschitz constant for A: {norm_x}")

    return norm_x



#!/usr/bin/python
import subprocess
from setuptools import setup, find_packages

with open("README.md", "r") as fh:
   long_description = fh.read()

# get all the git tags from the cmd line that follow our versioning pattern
git_tags = subprocess.Popen(
    ["git", "tag", "--list", "v*[0-9]", "--sort=version:refname"],
    stdout=subprocess.PIPE,
)
tags = git_tags.stdout.read()
git_tags.stdout.close()
tags = tags.decode("utf-8").split("\n")
tags.sort()

# PEP 440 won't accept the v in front, so here we remove it, strip the new line and decode the byte stream
VERSION_FROM_GIT_TAG = tags[-1][1:]

setup(
    name="surecr",
    version=VERSION_FROM_GIT_TAG,  # Required
    setup_requires=["setuptools>=18.0"],
    packages=find_packages(exclude=["notebooks"]),  # Required
    install_requires=[
        "numpy >= 1.17.5",
        "scipy",
        "torch",
        "cvxpy",
        "cvxpylayers",
        "torch-linops",
    ],
    description="This package enables tractable evaluation of Stein's Unbiased Risk Estimate on convexly regularized estimators.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/cvxgrp/SURE-CR",
    classifiers=[
        "Programming Language :: Python :: 3",
    ],
    author="Parth Nobel",
    author_email="ptnobel@stanford.edu",
)



import torch
from typing import Callable, Union
import sys

if sys.version_info >= (3, 10):
    floatTensor = float | torch.Tensor
else:
    floatTensor = Union[float,  torch.Tensor]
    

ProxOperator = Callable[[torch.Tensor, floatTensor], torch.Tensor]

def prox_l1_norm(v: torch.Tensor, t: floatTensor) -> torch.Tensor:
    """
    The ell_1 norm's proximal operator.
    """
    return torch.sign(v) * torch.relu(torch.abs(v) - t)

def make_scaled_prox_nuc_norm(shape: tuple[int, int], t_scale: floatTensor) -> \
        ProxOperator:
    """
    generates the proximal operator prox_r of r(b) = t_scale ||b||_*
    """
    t_scale = torch.as_tensor(t_scale)
    m, n = shape
    if m >= n:
        def prox_nuc_norm(x, t):
            X = x.reshape(shape)
            return _prox_nuc_norm.apply(X, t_scale * t).reshape(-1)
        return prox_nuc_norm
    else:
        def prox_nuc_norm_T(x, t):
            X = x.reshape(shape)
            return _prox_nuc_norm.apply(X.T, t_scale * t).T.reshape(-1)
        return prox_nuc_norm_T

def combine_proxs(shapes: list[int], proxs: list[ProxOperator]) -> ProxOperator:
    """
    Combines lists of proximal operators for seperable regularizers.

    If prox_r_1: R^a -> R^a and prox_r_2: R^b -> R^b then the arguments to this
    function should be [a, b], [prox_r_1, prox_r_2].
    """
    def combined_prox(v, t):
        total = 0
        out = torch.empty_like(v)
        for shape, prox in zip(shapes, proxs):
            mask = slice(total, total + shape)
            out[mask] = prox(v[mask], t)
            total += shape
        return out
    return combined_prox

def scale_prox(prox: ProxOperator, t_scale: floatTensor) -> ProxOperator:
    """
    Takes a proximal operator of r, and returns the proximal operator of t_scale r.
    """
    return lambda v, t: prox(v, t_scale * t)


def prox_l2_norm(v, t):
    """
    The ell_2 norm's proximal operator.
    """
    return torch.relu(1 - t / torch.linalg.norm(v)) * v

class _prox_nuc_norm(torch.autograd.Function):
    """
        Let X.shape == (m, n)
        This assumes m >= n.
    """
    @staticmethod
    #@torch.jit.script
    def forward(ctx, X, lambda_):
        def f(s):
            return torch.relu(s - lambda_)

        U, S, VT = torch.linalg.svd(X, full_matrices=True)
        ctx.U = U
        ctx.S = S
        ctx.VT = VT
        ctx.save_for_backward(lambda_)
        return U[:, :X.shape[1]] @ torch.diag(f(S)) @ VT

    @staticmethod
    #@torch.jit.script
    def backward(ctx, grad_output):
        epsilon = 1e-3
        lambda_, = ctx.saved_tensors
        U, S, VT = ctx.U, ctx.S, ctx.VT
        Z = grad_output
        m, n = Z.shape
        assert m >= n

        def f(s):
            return torch.relu(s - lambda_)
        def fp(s):
            #return (s >= lambda_).double()
            return (s >= lambda_).float()

        zeta = U.T @ Z @ VT.T
        Gamma = torch.empty_like(zeta)

        if m > n:
            R_mask = (S >= epsilon)
            R = torch.empty_like(S)
            R[R_mask] = f(S[R_mask]) / S[R_mask]
            R[~R_mask] = fp(S[~R_mask])
            Gamma[n:, :] = torch.tile(R, (m - n, 1)) * zeta[n:, :]

        S_i = torch.tile(S, (n, 1)).T 
        S_j = torch.tile(S, (n, 1))
        def off_diags_zeta(S_i, S_j):
            return (S_i * f(S_i) - S_j * f(S_j)) / (S_i**2 - S_j**2)
        def off_diags_zetaT(S_i, S_j):
            return (S_j * f(S_i) - S_i * f(S_j)) / (S_i**2 - S_j**2)
        zeta_weights = torch.empty_like(zeta[:n,:])
        zetaT_weights = torch.empty_like(zeta[:n,:])

        mask = torch.abs(S_i - S_j) <= epsilon

        zeta_weights[~mask] = off_diags_zeta(S_i[~mask], S_j[~mask])
        zetaT_weights[~mask] = off_diags_zetaT(S_i[~mask], S_j[~mask])

        def off_diags_pos_repeated_zeta(S):
            return (1/2 * fp(S) + 1/2 * f(S) / S)

        def off_diags_pos_repeated_zetaT(S):
            return (1/2 * fp(S) - 1/2 * f(S) / S)
        pos_mask = (mask & (S_i >= epsilon))
        zeta_weights[pos_mask] = off_diags_pos_repeated_zeta(S_i[pos_mask])
        zetaT_weights[pos_mask] = off_diags_pos_repeated_zetaT(S_i[pos_mask])

        zero_mask = (mask & (S_i < epsilon))
        zeta_weights[zero_mask] = fp(S_i[zero_mask])
        zetaT_weights[zero_mask] = 0

        torch.diagonal(zeta_weights)[:] = fp(S)
        torch.diagonal(zetaT_weights)[:] = 0

        Gamma[:n, :] = (
            zeta_weights * zeta[:n,:] + zetaT_weights * zeta[:n,:].T
        )

        retval = U @ Gamma @ VT
        return retval, None, None
